#pragma once
#include <cstdint>
#include <Redacted3D/types/entity/baseEntity.h>
#include <Redacted3D/types/vertex.h>
#include <Redacted3D/types/texture.h>
#include <J3ML/LinearAlgebra/Matrix4x4.hpp>
#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <Redacted3D/engine/collision.h>
//#include <archive.h>

using J3ML::LinearAlgebra::Matrix4x4;
using J3ML::LinearAlgebra::Vector3;

typedef Vector3 Direction;

///Base entity type.
class Entity : public BaseEntity {
private:
    ///Pitch Yaw Roll, The orientation of the entity in the world,
    Vector3 angle = {0,0,0};
    ///Loads the texture for the entity.

    ///The scale it should be rendered at.
    Vector3 scale = {1.0f, 1.0f, 1.0f};
protected:
///Loads the geometry for it.
void loadGeometry(const std::string& file);
void loadTexture(const std::string& file);
void loadMultiTexture(const std::vector<std::string>& files);
void loadMotionTexture(const std::vector<std::string>& files, u16 frameDelta, bool doAnim = true);

public:
    ///Whether an entity is solid. Entities are solid by default.
    bool collidable = true;

    ///Returns the texture for the entity. Loads it if it's not in the list.
    virtual Texture* getTexture();
    ///Returns the geometry for the entity. Loads it if it's not in the list.
    virtual VertexArray* getGeometry();
    ///Returns the AABB for the entity.
    AABB getAABB();
    ///Returns the minimal enclosing sphere for the entity.
    Sphere getSphere();
    ///Returns the OBB of the entity.
    OBB getOBB();
    void setScale(const float& multiplier);
    [[nodiscard]] Vector3 getScale() const;

protected:
    Matrix4x4 coordinates;
    J3ML::LinearAlgebra::Position position; /// X Y Z
public:
    u32 ticksAlive; //At 64tps it'd take 776 days to overflow.
    [[nodiscard]] Position GetPos() const;
    void SetPos(const Position& rhs);
    Matrix4x4 GetMatrix() const;
    void SetMatrix(const Matrix4x4& rhs);
    Direction getAngleDirection();
    Vector3 getAngle();
    void setAngle(const Vector3& a);
    void setAngle(float pitch, float yaw, float roll);
    ///Removes an entity from the list, Checks if the assets are being used by any other entity. Removes them from their lists & deletes. Then deletes the entity.
    void erase();
    bool occluded();
    bool isCollidingWith(Entity* entity);

    ///Default rendering routine. Works in 99% of cases.
    void render() override;

    Entity();

    Entity(const Entity& rhs) = default; // Boilerplate: Copy Constructor
    Entity(Entity&& rhs)      = default; // Boilerplate: Move Constructor
    ~Entity() = default;
};


