#include <Redacted3D/types/entity/baseEntity.h>
#include <stdexcept>

void BaseEntity::SetParent(BaseEntity* parent) {
    // hold a reference to this so it doesn't get collected as we're working with it
    BaseEntity* oldParent = this->parent;
    if (parent == oldParent)
        return;
    // Don't allow for an instance to be parented to itself
    if (this == parent)
        throw std::runtime_error("Cannot parent a widget to itself");
    if (this->IsAncestorOf(parent))
        throw std::runtime_error("Cannot create circular object reference");

    for (BaseEntity* ancestor: this->GetAncestors()) {
        if (oldParent && !ancestor->IsAncestorOf(parent) && parent != ancestor) {
            ancestor->DescendantRemoved(this);
            for (BaseEntity* descendant: this->GetDescendants()) {
                ancestor->DescendantRemoved(descendant);
            }
        }
    }

    // Remove ourselves from our parent (if we have one)
    if (this->parent) {
        for (auto it = parent->children.begin(); it != parent->children.end(); ++it) {
            if (*it == this) {
                parent->children.erase(it);
                break;
            }
        }
    }
    // Update our old parent to the new one
    this->parent = parent;
    // If our parent is set to nullptr, we can't update it's vector of children
    if (!parent) { return; }
    parent->children.emplace_back(this);
    for (BaseEntity* ancestor: this->GetAncestors()) {
        if (!oldParent || !(oldParent->IsDescendantOf(parent) && oldParent != ancestor)) {
            ancestor->DescendantAdded(this);
            for (auto* descendant : this->GetDescendants()) {
                ancestor->DescendantAdded(descendant);
            }
        }
    }
}

bool BaseEntity::IsDescendantOf(BaseEntity* ancestor) {
    if (ancestor == nullptr)
        return false;
    BaseEntity* instance = this;
    while (instance->GetParent()) {
        instance = instance->GetParent();
        if (instance == ancestor)
            return true;
    }
    return false;
}

bool BaseEntity::IsAncestorOf(BaseEntity* descendant) {
    if (descendant == nullptr)
        return false;
    BaseEntity* instance = descendant;
    while (instance->GetParent()) {
        instance = instance->GetParent();
        if (instance == this)
            return true;
    }
    return false;
}

std::vector<BaseEntity* > BaseEntity::GetChildren() {
    return this->children;
}

std::vector<BaseEntity* > BaseEntity::GetAncestors() {
    std::vector<BaseEntity* > ancestors;
    for (BaseEntity* ancestor = this->parent; ancestor; ancestor = ancestor->parent) {
        ancestors.push_back(ancestor);
    }
    return ancestors;
}

std::vector<BaseEntity* > BaseEntity::GetDescendants() {
    std::vector<BaseEntity* > descendants;
    for (auto& child: this->children) {
        descendants.push_back(child);
        std::vector<BaseEntity* > recursiveDescendants = child->GetDescendants();
        descendants.insert(descendants.end(), recursiveDescendants.begin(), recursiveDescendants.end());
    }
    return descendants;
}

BaseEntity* BaseEntity::Add(BaseEntity* newChild) {
    newChild->SetParent(this);
    return newChild;
}

BaseEntity* BaseEntity::GetFamilyTreeRoot() const {

    //if (JUI::Scene* scene = dynamic_cast<JUI::Scene*>(parent); scene != nullptr)

    // This is retarded, Fix ASAP!
    auto parent = this->GetParent();
    if (parent->GetParent() == nullptr)
        return parent;
    return parent->GetFamilyTreeRoot();
}